import pytest

from roam import r, r_strict, MISSING, Roamer, RoamPathException


class DataTester:
    """ Class to convert dict data to object with attributes """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        for n, v in kwargs.items():
            if n.startswith("_"):
                n = n[1:]
            setattr(self, n, v)

    def __dir__(self):
        return self.kwargs.keys()

    @property
    def as_dict(self):
        return self.kwargs


# Amended subset of GitHub user data for 'jmurty' from
# https://api.github.com/users/jmurty/repos
github_data = [
    {
        "name": "java-xmlbuilder",
        "full_name": "jmurty/java-xmlbuilder",
        "private": False,
        "owner": {
            "login": "jmurty",
            "url": "https://api.github.com/users/jmurty",
            "type": "User",
            "fn": lambda message: message,  # Added to test in-data callable
        },
        "description": "XML Builder is a utility that allows simple XML documents to be constructed using relatively sparse Java code",
        "fork": False,
        "url": "https://api.github.com/repos/jmurty/java-xmlbuilder",
        "created_at": "2014-03-05T22:48:04Z",
        "updated_at": "2019-04-29T14:03:24Z",
        "pushed_at": "2017-09-01T13:08:26Z",
        "homepage": None,
        "size": 164,
        "language": "Java",
        "archived": False,
        "disabled": False,
        "open_issues_count": 0,
        "license": {
            "key": "apache-2.0",
            "name": "Apache License 2.0",
            "spdx_id": "Apache-2.0",
            "url": "https://api.github.com/licenses/apache-2.0",
        },
        "forks": 15,
        "open_issues": 0,
        "watchers": 85,
        "default_branch": "master",
    },
    {
        "name": "xml4h",
        "full_name": "jmurty/xml4h",
        "private": False,
        "owner": {
            "login": "jmurty",
            "url": "https://api.github.com/users/jmurty",
            "type": "User",
        },
        "description": "XML for Humans in Python",
        "fork": False,
        "url": "https://api.github.com/repos/jmurty/xml4h",
        "created_at": "2012-08-14T13:49:43Z",
        "updated_at": "2018-09-21T16:15:44Z",
        "pushed_at": "2015-07-13T15:07:28Z",
        "homepage": "xml4h.readthedocs.org",
        "size": 663,
        "language": "Python",
        "archived": False,
        "disabled": False,
        "open_issues_count": 6,
        "license": {
            "key": "mit",
            "name": "MIT License",
            "spdx_id": "MIT",
            "url": "https://api.github.com/licenses/mit",
        },
        "forks": 2,
        "open_issues": 6,
        "watchers": 37,
        "default_branch": "master",
    },
]

github_data0 = github_data[0]


# Select data from https://en.wikipedia.org/wiki/Monty_Python_filmography
python_filmography = [
    DataTester(
        title="Monty Python's Flying Circus",
        type="tv",
        years=DataTester(_from=1975, to=1975),
        writers=[
            DataTester(name="Monty Python", group=True),
            DataTester(name="Neil Innes"),
            DataTester(name="Douglas Adams"),
        ],
    ),
    DataTester(
        title="Monty Python and the Holy Grail",
        type="movie",
        years=DataTester(_from=1969, to=1974),
        writers=[DataTester(name="Monty Python", group=True)],
    ),
]


class TestRoamer:
    def test_missing_has_rich_falsey_behaviour(self):
        assert not MISSING

        assert not MISSING

        assert len(MISSING) == 0

        for _ in MISSING:
            pytest.fail("Shouldn't be able to iterate over MISSING")

    def test_getattr_traversal(self):
        assert r(github_data0).license == github_data0["license"]

        assert r(github_data0).license.name == github_data0["license"]["name"]

        assert r(github_data0).license.name  # Truthy

    def test_attr_traversal_missing(self):
        assert r(github_data0).x == MISSING

        assert r(github_data0).license.x == MISSING

        assert not r(github_data0).license.x  # Falsey

        assert not r(github_data0).license.x.y  # Falsey

        # Confirm the underlying item is the MISSING singleton
        assert r(github_data0).license.x() is MISSING
        assert r(github_data0).license.x() == MISSING
        assert not r(github_data0).license.x()  # Falsey

    def test_getitem_traversal(self):
        assert r(github_data0)["license"] == github_data0["license"]

        assert r(github_data0)["license"]["name"] == github_data0["license"]["name"]

        assert r(github_data0).license.name[0] == github_data0["license"]["name"][0]

        assert r(github_data0).license.name[-1] == github_data0["license"]["name"][-1]

        assert r(github_data0).license.name[-1]  # Truthy

    def test_getitem_traversal_missing(self):
        assert r(github_data0)["x"] == MISSING

        assert r(github_data0)["license"]["x"] == MISSING

        assert r(github_data0)["license"]["name"]["x"] == MISSING

        assert not r(github_data0)["license"]["name"]["x"]  # Falsey

        assert not r(github_data0)["license"]["name"]["x"]["y"]  # Falsey

        # Confirm the underlying item is the MISSING singleton
        assert r(github_data0)["license"]["name"]["x"]() is MISSING
        assert r(github_data0)["license"]["name"]["x"]() == MISSING
        assert not r(github_data0)["license"]["name"]["x"]()  # Falsey

    def test_getattr_and_getitem_traversal(self):
        assert r(github_data0).license["name"] == github_data0["license"]["name"]

        assert r(github_data0)["license"].name == github_data0["license"]["name"]

        assert r(github_data0)["license"].name[1] == github_data0["license"]["name"][1]

    def test_getattr_and_getitem_traversal_missing(self):
        assert r(github_data0).license["x"] == MISSING

        assert r(github_data0)["license"].x == MISSING

        # Confirm the underlying item is the MISSING singleton
        assert r(github_data0).license["x"]() is MISSING
        assert r(github_data0)["license"].x() is MISSING

    def test_fail_fast(self):
        with pytest.raises(RoamPathException) as ex:
            r(github_data0, _raise=True).x
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 1 .x for path <dict>.x at <dict>"
            " with keys ['name', 'full_name', 'private', 'owner', 'description', 'fork',"
            " 'url', 'created_at', 'updated_at', 'pushed_at', 'homepage', 'size', 'language',"
            " 'archived', 'disabled', 'open_issues_count', 'license', 'forks', 'open_issues',"
            " 'watchers', 'default_branch']>"
        )

        with pytest.raises(RoamPathException) as ex:
            r(github_data0, _raise=True).license["x"]
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 2 ['x'] for path <dict>.license['x'] at <dict>"
            " with keys ['key', 'name', 'spdx_id', 'url']>"
        )

        with pytest.raises(RoamPathException) as ex:
            r(github_data0, _raise=True)["license"].name.x
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .x for path <dict>['license'].name.x at <str>>"
        )

    def test_slice_traversal(self):
        assert r(github_data)[:] == github_data[:]

        assert r(github_data)[1:] == github_data[1:]

    def test_slice_traversal_missing(self):
        assert len(github_data) == 2

        assert r(github_data)[3] == MISSING
        assert r(github_data)[3]() is MISSING

        assert r(github_data)[3:] == []

        assert r(github_data)[2:4] == github_data[2:4]

    def test_call_returns_item(self):
        assert r(github_data0).license.name() is github_data0["license"]["name"]

        assert r(github_data0)["license"]["name"]() is github_data0["license"]["name"]

    def test_call_returns_item_missing(self):
        assert r(github_data0).x() == MISSING
        assert r(github_data0)["x"]() is MISSING

        assert r(github_data0).license.x() == MISSING
        assert r(github_data0)["name"].x() is MISSING

    def test_call_raise_on_item_missing(self):
        with pytest.raises(RoamPathException) as ex:
            r(github_data0).x(_raise=True)
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 1 .x for path <dict>.x at <dict>"
            " with keys ['name', 'full_name', 'private', 'owner', 'description', 'fork',"
            " 'url', 'created_at', 'updated_at', 'pushed_at', 'homepage', 'size', 'language',"
            " 'archived', 'disabled', 'open_issues_count', 'license', 'forks', 'open_issues',"
            " 'watchers', 'default_branch']>"
        )

        with pytest.raises(RoamPathException) as ex:
            r(github_data0).license["name"].x(_raise=True)
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .x for path <dict>.license['name'].x at <str>>"
        )

        with pytest.raises(RoamPathException) as ex:
            r(github_data0, _raise=True).license["name"].x()
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .x for path <dict>.license['name'].x at <str>>"
        )

        # Test standard exception raised if user tries to call uncallable data
        with pytest.raises(TypeError) as ex:
            r(github_data0, _raise=True).license(1, 2, 3)
        assert str(ex.value) == "'dict' object is not callable"

        with pytest.raises(TypeError) as ex:
            r(github_data0, _raise=True).description(bad="argument")
        assert str(ex.value) == "'str' object is not callable"

    def test_call_delegates_to_and_returns_item(self):
        # Delegate to methods on `dict` item
        assert r(github_data0)["license"]["items"]() == github_data0["license"].items()
        assert r(github_data0).license.keys() == github_data0["license"].keys()
        assert list(r(github_data0).license.values()) == list(
            github_data0["license"].values()
        )

        # Delegate to methods on `str` item
        assert r(github_data0).license.url.split("/") == [
            "https:",
            "",
            "api.github.com",
            "licenses",
            "apache-2.0",
        ]
        assert type(r(github_data0).license.url.split("/")) is list

        # Delegate to callable item within traversal
        assert r(github_data0).owner.fn("Hi") == "Hi"
        assert r(github_data0).owner.fn(999) == 999

    def test_call_with_roam_option(self):
        assert isinstance(r(github_data0).license.items(_roam=True), Roamer)
        assert (
            r(github_data0).license.items(_roam=True) == github_data0["license"].items()
        )

    def test_call_on_missing_with_roam_option(self):
        assert isinstance(r(github_data0).x.items(_roam=True), Roamer)
        assert r(github_data0).x.items(_roam=True) == MISSING

    def test_call_with_invoke_option(self):
        assert r(github_data0).owner.login(_invoke=len) == 6

    def test_call_with_invoke_and_roam_options(self):
        assert isinstance(r(github_data0).owner(_invoke=len, _roam=True), Roamer)
        assert r(github_data0).owner(_invoke=len, _roam=True) == 4

    def test_iterator_traversal(self):
        for i, item_roamer in enumerate(r(github_data)):
            assert isinstance(item_roamer, Roamer)
            assert item_roamer.name == github_data[i]["name"]

        assert [
            owner_fn("Hello world!") for owner_fn in r(github_data)[:].owner.fn
        ] == ["Hello world!"]

        assert [
            owner.fn("Hello world!") for owner in r(github_data)[:].owner if owner.fn
        ] == ["Hello world!"]

        assert [
            writer.name()
            for writer in r(python_filmography)[:].writers
            if not writer.group
        ] == ["Neil Innes", "Douglas Adams"]

        # Trying to iterating over non-iterable
        for _ in r(github_data0).size:
            pytest.fail("Shouldn't be able to iterate over int")

        for _ in r(github_data0).fork:
            pytest.fail("Shouldn't be able to iterate over bool")

    def test_iterator_traversal_missing(self):
        for _ in r(github_data0).x:
            pytest.fail("Shouldn't be able to iterate over MISSING")

        for _ in r(github_data0).license.name.x:
            pytest.fail("Shouldn't be able to iterate over MISSING")

        with pytest.raises(RoamPathException) as ex:
            r_strict(github_data0).license.name.x
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .x for path <dict>.license.name.x at <str>>"
        )

    def test_nested_iterable_traversal(self):
        assert r(github_data)[:]["owner"]["login"] == ("jmurty", "jmurty")

        assert r(python_filmography)[:]["title"] == (
            "Monty Python's Flying Circus",
            "Monty Python and the Holy Grail",
        )
        assert r(python_filmography)[:]["writers"]["name"] == (
            "Monty Python",
            "Neil Innes",
            "Douglas Adams",
            "Monty Python",
        )
        assert r(python_filmography)[:].writers.name == (
            "Monty Python",
            "Neil Innes",
            "Douglas Adams",
            "Monty Python",
        )

        # Check slices work within multi-item processing
        assert (
            r(python_filmography)[:].writers.name[:]()
            == r(python_filmography)[:].writers.name()
        )
        assert (
            r(python_filmography)[:]["writers"]["name"][:]()
            == r(python_filmography)[:]["writers"]["name"]()
        )
        assert r(python_filmography)[:].writers.name[1:-1] == (
            "Neil Innes",
            "Douglas Adams",
        )

        # Check integer lookpus work within multi-item processing
        assert r(python_filmography)[:]["writers"][1]["name"] == "Neil Innes"

        assert r(python_filmography)[:].writers[1].name == "Neil Innes"

    def test_nested_iterable_traversal_missing(self):
        # Referencing missing attr/keys results in an empty list
        assert r(python_filmography)[:].x == tuple()
        assert r(python_filmography)[:]["x"] == tuple()

        assert r(python_filmography)[:].title.x == tuple()
        assert r(python_filmography)[:]["title"]["x"] == tuple()

        # Referencing *sometimes* missing attr/keys results in partial list
        assert len(r(python_filmography)[:].writers) == 4
        assert r(python_filmography)[:].writers.group == (True, True)
        assert r(python_filmography)[:]["writers"]["group"] == (True, True)

        # Lookup missing n-th item in a nested collection
        assert r(python_filmography)[:].writers.group[2] == MISSING
        with pytest.raises(RoamPathException) as ex:
            r(python_filmography)[:].writers.group[2](_raise=True)
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 4 [2] for path <list>[:].writers.group[2] at <tuple> with length 2>"
        )

    def test_roamer_equality(self):
        assert r(python_filmography)[:].writers == r(python_filmography)[:].writers

    def test_roamer_len(self):
        # Standard length lookup of list
        assert len(r(python_filmography)) == 2

        # Report 1 for current item that doesn't actually support `len()` instead of `TypeError`
        assert len(r(python_filmography)[0]) == 1
        with pytest.raises(TypeError):
            len(python_filmography[0])

        # Report zero from MISSING item
        x = r(python_filmography).x
        assert len(x) == 0
        assert x == MISSING

    def test_path_reporting(self):
        assert (
            str(r(github_data0).license.name)
            == "<Roamer: <dict>.license.name => 'Apache License 2.0'>"
        )

        assert (
            str(r(github_data0)["license"]["name"])
            == "<Roamer: <dict>['license']['name'] => 'Apache License 2.0'>"
        )

        assert (
            str(r(github_data0)["license"].name)
            == "<Roamer: <dict>['license'].name => 'Apache License 2.0'>"
        )

        assert (
            str(r(python_filmography)[:].writers[2]["as_dict"])
            == "<Roamer: <list>[:].writers[2]['as_dict'] => {'name': 'Douglas Adams'}>"
        )

        assert (
            str(r(python_filmography)[:].writers[2]["name"])
            == "<Roamer: <list>[:].writers[2]['name'] => 'Douglas Adams'>"
        )

        assert (
            str(r(python_filmography)[:].writers[3]["name"])
            == "<Roamer: <list>[:].writers[3]['name'] => 'Monty Python'>"
        )

        assert (
            str(r(python_filmography)[:].writers[2]["age"])
            == "<Roamer: missing step 4 ['age'] for path <list>[:].writers[2]['age'] at <DataTester> with attrs [name] => <MISSING>>"
        )

        assert (
            str(r(github_data0).license["x"])
            == "<Roamer: missing step 2 ['x'] for path <dict>.license['x'] at <dict>"
            " with keys ['key', 'name', 'spdx_id', 'url'] => <MISSING>>"
        )

        assert (
            str(r(python_filmography)[0].writers.name)
            == "<Roamer: missing step 3 .name for path <list>[0].writers.name at <list> => <MISSING>>"
        )

    def test_path_survives_roamer_reuse(self):
        roamer = r(github_data)

        assert (
            str(roamer[0].license.name)
            == "<Roamer: <list>[0].license.name => 'Apache License 2.0'>"
        )

        assert (
            str(roamer[1]["license"].name)
            == "<Roamer: <list>[1]['license'].name => 'MIT License'>"
        )

        with pytest.raises(RoamPathException) as ex:
            roamer[1].license.x(_raise=True)
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .x for path <list>[1].license.x at <dict> with keys ['key', 'name', 'spdx_id', 'url']>"
        )

        with pytest.raises(RoamPathException) as ex:
            roamer[1].license["name"].x(_raise=True)
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 4 .x for path <list>[1].license['name'].x at <str>>"
        )

    def test_roamer_clone_via_init(self):
        r_license = r(github_data)[0].license

        assert r_license == {
            "key": "apache-2.0",
            "name": "Apache License 2.0",
            "spdx_id": "Apache-2.0",
            "url": "https://api.github.com/licenses/apache-2.0",
        }

        # Ensure clone is identical in all ways that matter
        r_clone = r(r_license)
        assert r_clone == r_license
        assert r_clone._r_item_ == r_license._r_item_
        assert r_clone._r_is_multi_item_ == r_license._r_is_multi_item_
        assert r_clone._r_raise_ == r_license._r_raise_ is False

        # Can override raise status in clone
        r_clone = r(r_license, _raise=True)
        assert r_clone._r_raise_ is True
        with pytest.raises(RoamPathException) as ex:
            r_clone.wrong
        assert (
            str(ex.value)
            == "<RoamPathException: missing step 3 .wrong for path <list>[0].license.wrong"
            " at <dict> with keys ['key', 'name', 'spdx_id', 'url']>"
        )



""" Easily traverse nested Python data structures """

__version__ = "0.3.1"


class _RoamMissingItem:
    """ Falsey class used to flag item "missing" from traversal path """

    def __bool__(self):
        return False

    def __len__(self):
        return 0

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration()

    def __repr__(self):
        return "<MISSING>"


MISSING = _RoamMissingItem()


class _Path:
    _r_root_item_ = None
    _r_steps_ = []

    def __init__(self, initial_item, path_to_clone=None):
        if path_to_clone is not None:
            self._r_root_item_ = path_to_clone._r_root_item_
            self._r_steps_ = list(path_to_clone._r_steps_)  # Shallow copy list
        else:
            self._r_root_item_ = initial_item
            self._r_steps_ = []

    def log_getattr(self, attr_name: str, roamer: "Roamer"):
        """
        Log the fact that a ``.dot`` attribute lookup was performed using a
        given name and the given ``Roamer`` shim was produced.
        """
        self._r_steps_.append((f".{attr_name}", unwrap(roamer)))

    def log_getitem(self, slice_value: slice, roamer: "Roamer"):
        """
        Log the fact that a ``["slice"]`` attribute lookup was performed using a
        given slice value and the given ``Roamer`` shim was produced.
        """
        if isinstance(slice_value, slice):
            item_desc = (
                f"[{slice_value.start or ''}:{slice_value.stop or ''}"
                f"{slice_value.step and ':' + slice_value.step or ''}]"
            )
        else:
            item_desc = f"[{slice_value!r}]"
        self._r_steps_.append((item_desc, unwrap(roamer)))

    def _last_found(self):
        last_found_step = None, None, self._r_root_item_
        for i, step in enumerate(self._r_steps_, 1):
            desc, data = step
            if data is not MISSING:
                last_found_step = i, desc, data
        return last_found_step

    def _first_missing(self):
        for i, step in enumerate(self._r_steps_, 1):
            desc, data = step
            if data is MISSING:
                return i, desc, data
        return None, None, self._r_root_item_

    def description(self) -> str:
        """
        Return a text description of this path, capturing:
        - the first step at which the path was invalid (if applicable)
        - the type of the root data object
        - path steps applied
        - hints about the type and content of data at the point the path became
          invalid (if applicable)
        """
        result = []

        first_missing_index, first_missing_desc, _ = self._first_missing()
        if first_missing_index:
            result.append(
                f"missing step {first_missing_index} {first_missing_desc} for path "
            )

        result.append(f"<{type(self._r_root_item_).__name__}>")
        result += [desc for desc, _ in self._r_steps_]

        if first_missing_index:
            _, _, last_found_data = self._last_found()
            if last_found_data is not MISSING:
                result.append(f" at <{type(last_found_data).__name__}>")

                # Generate hints
                if isinstance(last_found_data, (tuple, list, set, range)):
                    # Detect an integer key slice operation like `[3]` or `[-2]`
                    if first_missing_desc[0] == "[" and first_missing_desc[-1] == "]":
                        try:
                            int(first_missing_desc[1:-1])
                            result.append(f" with length {len(last_found_data)}")
                        except ValueError:
                            pass
                elif isinstance(
                    last_found_data, (str, int, float, complex, bool, bytes, bytearray)
                ):
                    pass  # No hint for primitive types
                elif last_found_data:
                    try:
                        keys = last_found_data.keys()
                        if keys:
                            result.append(
                                f" with keys [{', '.join([repr(k) for k in keys])}]"
                            )
                    except AttributeError:
                        attrs = dir(last_found_data)
                        if attrs and not isinstance(
                            last_found_data, (str, tuple, list)
                        ):
                            result.append(
                                f" with attrs [{', '.join([a for a in attrs if not a.startswith('_')])}]"
                            )

        return "".join(result)

    def __eq__(self, other):
        if isinstance(other, _Path):
            return (
                self._r_root_item_ == other._r_root_item_
                and self._r_steps_ == other._r_steps_
            )
        return False


class RoamPathException(Exception):
    """
    An exception raised when a ``Roamer`` shim encounters an invalid path step
    if that shim has the ``_raise`` option set, or provided when returning data.

    The ``str()`` representation of this exception is a rich description of
    where your traversal path went wrong.
    """

    def __init__(self, path):
        super().__init__(self)
        self.path = path

    def __str__(self):
        return f"<RoamPathException: {self.path.description()}>"


class Roamer:
    """
    Act as a shim over your data objects, to intercept Python operations and do
    the extra work required to more easily traverse nested data.
    """

    # Internal state variables
    _r_item_ = None
    _r_path_ = None
    _r_is_multi_item_ = False
    # Options
    _r_raise_ = False
    # Temporary flags
    _r_via_alternate_lookup_ = False
    _r_item__iter = None

    def __init__(self, item, _raise=None):
        # Handle `item` that is itself a `Roamer`
        if isinstance(item, Roamer):
            for attr in ("_r_item_", "_r_is_multi_item_", "_r_raise_"):
                setattr(self, attr, getattr(item, attr))
            self._r_path_ = _Path(item._r_item_, item._r_path_)
        else:
            self._r_item_ = item
            self._r_path_ = _Path(self._r_item_)
        # Set or override raise flag if user provided a value
        if _raise is not None:
            self._r_raise_ = bool(_raise)

    def __getattr__(self, attr_name):
        # Stop here if no item to traverse
        if self._r_item_ is MISSING:
            if not self._r_via_alternate_lookup_:
                self._r_path_.log_getattr(attr_name, self)
            return self

        copy = Roamer(self)
        # Multi-item: `.xyz` => `(i.xyz for i in item)`
        if self._r_is_multi_item_:
            multi_items = []
            for i in self._r_item_:
                lookup = None
                try:
                    lookup = getattr(i, attr_name)
                except (TypeError, AttributeError):
                    try:
                        lookup = i[attr_name]
                    except (TypeError, LookupError):
                        pass
                if isinstance(lookup, (tuple, list, range)):
                    multi_items += lookup
                elif lookup is not None:
                    multi_items.append(lookup)
            copy._r_item_ = tuple(multi_items)
        # Single item: `.xyz` => `item.xyz`
        else:
            try:
                copy._r_item_ = getattr(copy._r_item_, attr_name)
            except (TypeError, AttributeError):
                # Attr lookup failed, no more attr lookup options
                copy._r_item_ = MISSING

        # Fall back to `self.__getitem__()` if lookup failed so far and we didn't come from there
        if copy._r_item_ is MISSING and not self._r_via_alternate_lookup_:
            try:
                self._r_via_alternate_lookup_ = True
                copy = self[attr_name]
            except RoamPathException:
                pass
            finally:
                copy._r_path_.log_getattr(attr_name, copy)
                self._r_via_alternate_lookup_ = False
        elif not self._r_via_alternate_lookup_:
            copy._r_path_.log_getattr(attr_name, copy)

        if copy._r_item_ is MISSING and copy._r_raise_:
            raise RoamPathException(copy._r_path_)

        return copy

    def __getitem__(self, key_or_index_or_slice):
        # Stop here if no item to traverse
        if self._r_item_ is MISSING:
            if not self._r_via_alternate_lookup_:
                self._r_path_.log_getitem(key_or_index_or_slice, self)
            return self

        copy = Roamer(self)
        # Multi-item: `[xyz]` => `(i[xyz] for i in item)`
        if copy._r_is_multi_item_ and not isinstance(key_or_index_or_slice, slice):
            # Flatten item if we have selected a specific integer index
            if isinstance(key_or_index_or_slice, int):
                try:
                    copy._r_item_ = copy._r_item_[key_or_index_or_slice]
                except (TypeError, LookupError):
                    copy._r_item_ = MISSING
                # No longer in a multi-item if we have selected a specific index item
                copy._r_is_multi_item_ = False
            # Otherwise apply slice lookup to each of multiple items
            else:
                multi_items = []
                for i in copy._r_item_:
                    lookup = None
                    try:
                        lookup = i[key_or_index_or_slice]
                    except (TypeError, LookupError):
                        try:
                            lookup = getattr(i, key_or_index_or_slice)
                        except (TypeError, AttributeError):
                            pass
                    if isinstance(lookup, (tuple, list, range)):
                        multi_items += lookup
                    elif lookup is not None:
                        multi_items.append(lookup)
                copy._r_item_ = tuple(multi_items)
        # Lookup for non-multi item data, or for slice lookups in all cases
        else:
            try:
                copy._r_item_ = copy._r_item_[key_or_index_or_slice]
            except (TypeError, LookupError):
                # Index lookup failed, no more index lookup options
                copy._r_item_ = MISSING

        # Flag the fact our item actually has multiple elements
        if isinstance(key_or_index_or_slice, slice):
            copy._r_is_multi_item_ = True

        # Fall back to `self.__getattr__()` if lookup failed so far and we didn't come from there
        if (
            copy._r_item_ is MISSING
            and not self._r_via_alternate_lookup_
            # Cannot do an integer attr lookup
            and not isinstance(key_or_index_or_slice, int)
        ):
            try:
                self._r_via_alternate_lookup_ = True
                copy = getattr(self, key_or_index_or_slice)
            except RoamPathException:
                pass
            finally:
                copy._r_path_.log_getitem(key_or_index_or_slice, copy)
                self._r_via_alternate_lookup_ = False
        elif not self._r_via_alternate_lookup_:
            copy._r_path_.log_getitem(key_or_index_or_slice, copy)

        if copy._r_item_ is MISSING and copy._r_raise_:
            raise RoamPathException(copy._r_path_)

        return copy

    def __call__(self, *args, _raise=False, _roam=False, _invoke=None, **kwargs):
        if _raise and self._r_item_ is MISSING:
            raise RoamPathException(self._r_path_)

        # If an explicit callable is provided, call `_invoke(item, x, y, z)`
        if _invoke is not None:
            call_result = _invoke(self._r_item_, *args, **kwargs)
        # If item is callable: `.(x, y, z)` => `item(x, y, z)`
        elif callable(self._r_item_):
            call_result = self._r_item_(*args, **kwargs)
        # If item is not callable but we were given parameters, try to apply
        # them even though we know it won't work, to generate the appropriate
        # exception to let the user know their action failed
        elif args or kwargs:
            call_result = self._r_item_(*args, **kwargs)
        # If item is not callable: `.()` => return wrapped item unchanged
        else:
            call_result = self._r_item_

        # Re-wrap return as a `Roamer` if requested
        if _roam:
            copy = Roamer(self)
            copy._r_item_ = call_result
            return copy
        return call_result

    def __iter__(self):
        try:
            self._r_item__iter = iter(self._r_item_)
        except (TypeError, AttributeError):
            self._r_item__iter = None
        return self

    def __next__(self):
        if self._r_item__iter is None:
            raise StopIteration()
        next_value = next(self._r_item__iter)
        return Roamer(next_value)

    def __eq__(self, other):
        if isinstance(other, Roamer):
            for attr in ("_r_item_", "_r_path_", "_r_is_multi_item_", "_r_raise_"):
                if getattr(other, attr) != getattr(self, attr):
                    return False
            return True
        else:
            return other == self._r_item_

    def __bool__(self):
        return bool(self._r_item_)

    def __len__(self):
        try:
            return len(self._r_item_)
        except TypeError:
            # Here we know we have a non-MISSING item, but it doesn't support length lookups so
            # must be a single thing...
            # WARNING: This is black magic, does it make enough sense?
            return 1

    def __repr__(self):
        return f"<Roamer: {self._r_path_.description()} => {self._r_item_!r}>"


def r(item: object, _raise: bool = None) -> Roamer:
    """
    A shorter alias for constructing a ``Roamer`` shim class.
    """
    return Roamer(item, _raise=_raise)


def r_strict(item: object) -> Roamer:
    """
    A shorter alias for constructing a ``Roamer`` shim class in "strict" mode,
    which means that the ``_raise`` flag set so the shim will immediately raise
    a ``RoamPathException`` when you express an invalid path step.
    """
    return Roamer(item, _raise=True)


def unwrap(roamer: Roamer, _raise: bool = None) -> object:
    """
    Return the underlying data in the given ``Roamer`` shim object without
    the need to call that shim object.

    This is not the recommended way to get data from **roam** but you might
    prefer it, or it might help to solve unexpected bugs caused by the semi-
    magical call behaviour.
    """
    result = roamer._r_item_
    if _raise and result is MISSING:
        raise RoamPathException(roamer._r_path_)
    return result



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='roam',
      version='0.3.1',
      description='Easily traverse nested Python data structures ',
      author='James Murty',
      author_email='james@murty.co',
      url='https://github.com/jmurty/roam/',
      py_modules=['roam'],
      python_requires='>=3.6',
     )
